//file keypad4x4.c
//read 4x4 keypad on portc
//for ere co embmega32 16Mhz 38400

//Jan 9 11 Bob G 

#include <iom32v.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <macros.h>

#define INTR_OFF() asm("CLI")
#define INTR_ON()  asm("SEI")

//------vars in bss-------
unsigned int tics,ticsl;

//------rom---------
__flash unsigned char banner[]={"keypad4x4 Jan 9 11\n"};

//-----------------------
void port_init(void){

  PORTA = 0x00; //pullups off
  DDRA  = 0x00; //all ins
 
  PORTB = 0xFF; 
  DDRB  = 0xFF; 
 
  PORTC = 0x0F; //pullups on lo nib
  DDRC  = 0xF0; //hi nib outs
 
  PORTD = 0xFF; //
  DDRD  = 0xFF; //
}

#if 0
//------------------
//SPI initialisation
// clock rate: 115200hz
void spi_init(void){
  SPCR = 0x51; //setup SPI (slowest clock rate! 0x50 fastest)
  SPSR = 0x00; //setup SPI
}
#endif

//--------------------
//UART0 initialize
// desired baud rate: 38400
// actual: baud rate:38462 (0.2%)
// char size: 8 bit
// parity: Disabled
void uart0_init(void){
  UCSRB = 0x00; //disable while setting baud rate
  UCSRA = 0x00;
  UCSRC = BIT(URSEL) | 0x06;
  UBRRL = 0x19; //set baud rate lo
  UBRRH = 0x00; //set baud rate hi
  UCSRB = 0x18;
}

#if 0
//--------------------------
//ADC initialisation
// Conversion time: 112uS
void adc_init(void){
  ADCSRA = 0x00; //disable adc
  ADMUX  = 0x00; //select adc input 0
  ACSR   = 0x80;
  ADCSRA = 0x87;
}

//-----------------------
//TIMER0 initialize - prescale:8
// desired value: 16KHz
// actual value:  16.000KHz (0.0%)
void timer0_init(void){
  TCCR0 = 0x00; //stop
  TCNT0 = 0x83; //set count
  OCR0  = 0x7d;  //set compare
  TCCR0 = 0x0A; //start timer
}
#endif

#if 0
//------------------------
//TIMER1 initialisation - prescale:1
// WGM: 0) Normal, TOP=0xFFFF
// desired value: 16000KHz
// actual value: 16000.000KHz (0.0%)
void timer1_init(void){
  TCCR1B = 0x00; //stop
  TCNT1H = 0xFF; //setup
  TCNT1L = 0xFF;
  OCR1AH = 0x00;
  OCR1AL = 0x01;
  OCR1BH = 0x00;
  OCR1BL = 0x01;
  ICR1H  = 0x00;
  ICR1L  = 0x01;
  TCCR1A = 0x80;
  TCCR1B = 0x01; //start Timer
}
#endif

//-----------------------
void init_devices(void){
  CLI(); //disable all interrupts
  port_init();
//  spi_init();
  uart0_init();
//  adc_init();
//  timer0_init(); //8KHz sample rate clk
//  timer1_init(); //16 KHz pwm

  MCUCR = 0x00; //no ext mem
//  TIMSK = 0x14; //timer interrupt sources   t0 compare 
//  SEI(); //re-enable interrupts
}

//-------------------
void delnms(int n){
//delay n ms
int x;

  while(n--){
//    x=2400;       //empirically determined fudge factor  14.7456 mhz
    x=2600;       //empirically determined fudge factor  16 mhz
//    x=3000;       //empirically determined fudge factor  18.432 mhz
    while(x--);
  }
}

//-------------------------
void initvars(void){
//init vars

}

//-------RS232------------------
unsigned char kbhit(void){
//return nonzero if char waiting  polled version
unsigned char b;

  b=0;
  if(UCSRA & (1<<RXC)) b=1;
  return b;
}

//-----------------------
int getchar(void){
//polled version... 
unsigned char c;

  while(!kbhit()){}; //wait for char
  c=UDR;             //get char from usart
  return c;  
}

//-----------------------
int putc(char c){
//put char to usart0... dont add lf after cr

  while((UCSRA & 0x20)==0){};	//0x20 is UDRE0. ==0 means data register full
  UDR=c; //send char
  return c;
}

//----------------------  
int putchar(char c)	{
//adds lf after cr

  if(c=='\n'){
    putc('\r');
  }  
  putc(c);
  return c;
}

//-----------------
void crlf(void){
//output cr then lf

  putchar('\n');
}
  
//-----------------
void space(void){
//output space

  putc(' ');
}
  
//-----------------
void spaces(char n){
//output n spaces

  while(n--)
    space();
}
  
#define ESC 0x1b
//----------------------  
void gotoxy(int x, int y){
//ansi cursor positioning sequence E[y;xH

  putc(ESC);
  putc('[');
  cprintf("%d",y);
  putc(';');
  cprintf("%d",x);
  putc('H');
}

//-------------------------
unsigned char getche(void){
//get and echo a char from rs232
char c;

  c=getchar();
  putchar(c);
  return(c);
}

//----------------------
void clrscr(void){
//clear ansi terminal screen

  putc(ESC);
  putc('[');
  putc('2');
  putc('J');
}

//------------------
void initscreen(void){
//clear screen and print banner

  clrscr();
  cprintf(banner);
}

//-------------------------
unsigned char gethex(void){
//return a hex char from rs232
unsigned char b,c;

  b=0xff; //error return value
  c=toupper(getche());
  if(isxdigit(c)) b=c-0x30; //if c ok, cvt ascii digit to binary
  if((c>='A') && (c<='F')) b-=7;     //if c hexcvt ascii A to binary 10
  return(b);
}

//---------------------------
unsigned char getbyte(void){
//get 2 nibbles, return a binary byte from rs232
unsigned char n1,n2;

  n1=gethex();
  n2=gethex();
  return((n1 << 4) + n2);
}

//----------------------------
unsigned int getaddr(void){
//return addr from rs232
unsigned int th,tl;

  th=getbyte();
  tl=getbyte();
  return((th << 8) + tl);
}

//-----------------
void getstr(char *p){
//get a string and echo
char c;  

  do{
    c=getche();
    *p++=c;
  }while(c !=0x0d);
  p--;     //roll p back
  *p=0x00; //add null over cr
}

//------debug subs-----------
void dump256ram(unsigned char *n){
//dump 16 rows of 16 bytes
unsigned char r,c,ch;
unsigned char *p;
unsigned char *pa;

  p=n;
  crlf();
  spaces(6);
  for(c=0; c<16; c++){
    cprintf("%02x ",c);      //header
    if(c==7) space();
  }
  crlf();
  crlf();
  for(r=0; r<16; r++){
    cprintf("%04x  ",p);     //print addr  at beg of line
    pa=p;            //remember p for ascii
    for(c=0; c<16; c++){
      cprintf("%02x ",*p++); //print hex
      if(c==7) space();
    }
    for(c=0; c<16; c++){
      ch=*pa++;
      if((ch > 0x20) && (ch !=0x0a) && (ch != 0x8a)) //if printing char
        putc(ch);   //print ascii
      else  
        putc('.'); 
      if(c==7) space();
    }
    crlf();
  }
}

//-------------------------
void dump256rom(__flash unsigned char *n){
//dump 16 rows of 16 bytes
unsigned char r,c,ch;
__flash unsigned char *p;
__flash unsigned char *pa;

  p=n;
  crlf();
  spaces(6);
  for(c=0; c<16; c++){
    cprintf("%02x ",c);      //header
    if(c==7) space();
  }
  crlf();
  crlf();
  for(r=0; r<16; r++){
    cprintf("%04x  ",p);     //print addr  at beg of line
    pa=p;            //remember p for ascii
    for(c=0; c<16; c++){
      cprintf("%02x ",*p++); //print hex
      if(c==7) space();
    }
    for(c=0; c<16; c++){
      ch=*pa++;
      if((ch > 0x20) && (ch !=0x0a) && (ch != 0x8a)) //if printing char
        putc(ch);   //print ascii
      else  
        putc('.'); 
      if(c==7) space();
    }
    crlf();
  }
}

//------------------
void examineram(void){
//ask for mem range in ram and dump
unsigned char *from;
unsigned char c;

  cprintf("from:");
  from=(unsigned char *)getaddr();
  while(c!='q'){
    dump256ram(from);
    cprintf("np or q...");
    c=getchar();
    if(c=='n') from+=0x100;
    if(c=='p') from-=0x100;
  }  
}

//------------------
void examinerom(void){
//ask for mem range in rom and dump
__flash unsigned char *from;
unsigned char c;

  cprintf("from:");
  from=(__flash unsigned char *)getaddr();
  while(c!='q'){
    dump256rom(from);
    cprintf("np or q...");
    c=getchar();
    if(c=='n') from+=0x100;
    if(c=='p') from-=0x100;
  }  
}

//------------------
void deposit(void){
//ask for addr and data
unsigned char *at;
unsigned char c;
unsigned char nh,nl;

  cprintf("at:");
  at=(unsigned char *)getaddr();
  while(1){
    cprintf(" %02x ",*at);
    nh=gethex();
    if(nh==0xff) return;
    nl=gethex();
    if(nl==0xff) return;
    c=((nh << 4) | nl);
    *at++=c;
  }
}

//------------------
void fill(void){
//ask for mem range and fill char and fill
unsigned char *from, *to, with;

  cprintf("from:");
  from=(unsigned char *)getaddr();
  cprintf(" to:");
  to=(unsigned char *)getaddr();
  cprintf(" with:");
  with=getbyte();
  memset(from,with,to-from);
}

#if 0
//-----------------------------
void (* fn)(void); //fn prototype

void dojsr(void){
//ask for addr, jsr to it
__flash unsigned char *to;

  cprintf(" to:");
  to=(__flash unsigned char *)getaddr();
  fn=(void *)to;
  (*fn)();  //call function fn and returns
}
#endif

//------------------
void debugmenu(void){
//monitor cmds

  cprintf("cmds:\n" 
          " e examine ram\n"
          " c examine rom\n"
          " d deposit\n"
          " f fill\n"
//          " j jsr\n"
          " q quit\n");
}

//------------------
void debugloop(void){
//examine and deposit mem regs
char c;

  debugmenu();
  while(c != 'q'){
    crlf();
    cprintf("cmd:");
    c=getchar();
    crlf();
    switch(c){
      case 'e': examineram(); break;
      case 'c': examinerom(); break;
      case 'd': deposit(); break;
      case 'f': fill();    break;
//      case 'j': dojsr();   break;
      default: debugmenu();
    }
  }
}
//----end of debug subs----------

//---------------------
void initkeypad(void){
//make portc  hi nib outs, lo nib ins,

  DDRC = 0xf0;
}

//----------------------
char anykey(void){
//return true if key on pad pressed
char c,rowcode;

  PORTC &= 0x0F; //all outs lo (hi nib)
  delnms(1);
  rowcode=PINC & 0x0F;  //read row inputs.. should be ff or 7,b,d,e
  c=(rowcode & 0x0F) != 0x0F;
//  cprintf("        %02x %02x\n",c,rowcode);
  PORTC |= 0xF0; //all outs off
  return(c);
}

char ak,akl,akos;
//----------------------
char kphit(void){
//return true if key pad pressed oneshot

  ak=anykey();
  akos=ak && !akl;
  akl=ak;
  return akos;  
}

//-------------------
__flash char keytab[16]={
// 0   1   2   3 <-col
  'C','D','E','F', //3
  '8','9','A','B', //2
  '4','5','6','7', //1
  '0','1','2','3'  //0 <-row
};

//--------------------
char getkp(void){
//return key pressed on pad     outs in hi nib
char c,row,col,rowcode,keyndx;

  for(col=0; col < 4; col++){
    PORTC &= ~(0x80 >> col);   //out 0,1,2,3 lo
    delnms(1);
    rowcode=PINC & 0x0F;       //7,b,d,e=row 0,1,2,3
    PORTC |= 0xff;             //all outs off
 //   cprintf("     %02x\n",rowcode);
    if(rowcode != 0x0F) break; //this col has a row lo
  }
  if(rowcode == 0x0F){
    c=0xff;
  }else{  
    switch(rowcode){
      case 0x07: row=3; break;
      case 0x0b: row=2; break;
      case 0x0d: row=1; break;
      case 0x0e: row=0; break;
    }//switch
    keyndx=col*4 + row;
    c=keytab[keyndx];
  }  
//  cprintf("%02x %02x %02x %02x\n",rowcode,row,col,keyndx);
  return(c);
}

//--------------------
void keypad(void){
//read keypad
char c,d;

  while(c != 'q'){
    if(kbhit()){
      c=getchar();
    }
    if(kphit()){
      d=getkp();
      putc(d);
    }  
    delnms(20);
  }  
}

//----------------
void raw(void){
//raw port b
char c,d;

  cprintf("raw port data  az=00/ff s=1 x=2 d=4 c=8  q to quit\n");
  d=0;
  while(c != 'q'){
    if(kbhit()){
      c=getchar();
      switch(c){
        case 'a': d=0x00; break;
        case 'z': d=0xff; break;
        case 's': d= ~0x01; break;
        case 'x': d= ~0x02; break;
        case 'd': d= ~0x04; break;
        case 'c': d= ~0x08; break;
      }
      PORTC=d;
    }//if
    cprintf("%02x \r",PINC);
    delnms(20);
  }//while  
}

//--------------------------
void menu(void){
//cmds

  cprintf("cmds:\n");
  cprintf(" k keypad\n");
  cprintf(" r raw\n");
  cprintf(" d debug\n");
}

//-----------------
void main(void){
//main program
char c;

  init_devices();
  initvars();
  initscreen(); //banner
  while(1){     //main loop
    menu();
    crlf();
    putc('>'); //prompt for input
    c=getche();
    crlf();
    switch(c){
      case 'k': keypad(); break;
      case 'r': raw(); break;
      case 'd': debugloop();  break;
      default: menu();
    }
  }
}

//---strings embedded in rom--------------
__flash char copyrightstring[]={"keypad4x4 computer code\
written by Bob Gardner 407-855-2586 bobgardner at aol.com"};
//--------eof-------------
